package net.sergeych.crypto2

import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.bintools.CRC
import net.sergeych.bintools.CRC8
import net.sergeych.mp_tools.decodeBase64Url
import kotlin.random.Random

@Serializable
class BinaryId(
    val id: UByteArray,
) : Comparable<BinaryId> {

    class InvalidException(text: String) : IllegalArgumentException(text)
    class IncomparableException(text: String) : IllegalArgumentException(text)

    @Transient
    val magick: Int = run {
        if (id.size < 4) throw InvalidException("BinaryId is too short")
        val crc = id.last()
        val rest = id.dropLast(1).toUByteArray()
        if (CRC.crc8(rest) != crc)
            throw InvalidException("Bad BinaryId CRC")

        rest.last().toInt()
    }

    private val innerData: UByteArray by lazy { id.sliceArray( 1..< id.size-1 ) }

    override fun toString(): String = id.encodeToBase64Url()

    override fun compareTo(other: BinaryId): Int {
        if (other.magick != magick) throw IncomparableException("Mask mismatch (my=$magick their=${other.magick})")
        val id1 = other.id
        if (id1.size != id.size) throw IncomparableException("different sizes")
        for ((a, b) in innerData.zip(other.innerData)) {
            if (a < b) return -1
            if (a > b) return 1
        }
        return 0
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other !is BinaryId) return false

        if (!id.contentEquals(other.id)) return false

        return true
    }

    override fun hashCode(): Int {
        return id.contentHashCode()
    }


    companion object {

        @Suppress("unused")
        fun fromString(str: String): BinaryId =
            BinaryId(str.decodeBase64Url().toUByteArray())


        fun createFromBytes(magick: Int, bytes: ByteArray): BinaryId = createFromUBytes(magick, bytes.toUByteArray())

        fun createFromUBytes(magick: Int, bytes: UByteArray): BinaryId {
            val crc = CRC8()
            val mn = magick.toUByte()
            crc.update(bytes)
            crc.update(mn)
            return BinaryId(UByteArray(bytes.size + 2).also {
                bytes.copyInto(it, 0)
                val n = bytes.size
                it[n] = mn
                it[n + 1] = crc.value
            })
        }

        @Suppress("unused")
        fun createRandom(magickNumber: Int, size: Int=16) =
            createFromBytes(magickNumber, Random.Default.nextBytes(size-2))

    }
}